/*
    append - creates a compilation of goomba v2.23 compliant borders for 
             appending/prepending
    Copyright (C) 2005  Kuwanger

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <X11/Xlib.h>
#include <Imlib2.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>

int LZ_Compress( unsigned char *in, unsigned char *out, unsigned int insize ); // from lz.c
typedef unsigned char u8;
typedef unsigned short u16;

typedef struct {
    u8 b;
    u8 g;
    u8 r;
    u8 a;
} color;

typedef struct {
    u8 raw[32];
} agb_font;

#define MAXCOLORS 16	// 0 is transparent
#define MAXFONTS 216    // Only enough space for ~1/3 the screen to have unique fonts
#define ADDER 0xE040    // (Palette used)<<8 + (font start address)>>5

const unsigned int borderid=0x789b4987;

unsigned short agb_rgb(DATA32 * n)
{
    color *c = (color *) n;
    u16 col =
	((c->
	  r >> 3) & 31) | (((c->g >> 3) & 31) << 5) | (((c->
							 b >> 3) & 31) <<
						       10);
    return (col == 0) ? (1 | (1 << 5) | (1 << 10)) : col;
}

u8 updatepalette(unsigned short *p, u8 pidx, DATA32 * n)
{
    u8 tidx;

    p[pidx] = agb_rgb(n);


    for (tidx = 1; tidx < pidx; tidx++) {
	if (p[pidx] == p[tidx])
	    return pidx;
    }

    return pidx + 1;
}

u8 frompalette(unsigned short *p, u8 pidx, DATA32 * n)
{
    int i;
    unsigned short temp = agb_rgb(n);

    for (i = 1; i < pidx; i++) {
	if (p[i] == temp)
	    return i;
    }
    for (i = 1; i < pidx; i++) {
	printf("%d", p[i]);
	if (i == pidx - 1)
	    printf("\n");
	else
	    printf(", ");
    }
    printf("%d\n", temp);
    printf("Error. No match found from palette.\n");
    exit(1);
}

agb_font *horizontal_mirror(agb_font * font)
{
   agb_font *f = malloc(sizeof(agb_font));
   int i,j;
   
    for (j = 0; j < 8; j++) {
	for (i = 0; i < 4; i ++) {
	   f->raw[j*4+(3-i)] = ((font->raw[j * 4 + i]&0xf)<<4)|((font->raw[j * 4 + i]>>4)&0xf);
	}
    }
    return f;
}

agb_font *vertical_mirror(agb_font * font)
{
   agb_font *f = malloc(sizeof(agb_font));
   int i,j;
   
    for (j = 0; j < 8; j++) {
	for (i = 0; i < 4; i ++) {
	   f->raw[(7-j)*4+i] = font->raw[j * 4 + i];
	}
    }
    return f;
}

/* Add tile pointed at by in to font, if it's new
 * and point to where to add the next font.  Note,
 * each vertical line is 240 below the last.
 */
u8 updatefont(agb_font * font, u8 pos, unsigned short *palette, u8 * pal,
	      u16 * screen, DATA32 * in)
{
    agb_font *v,*h,*vh;
    u8 temp;
    int i, j;
    for (j = 0; j < 8; j++) {
	for (i = 0; i < 8; i += 2) {
	    *pal = updatepalette(palette, *pal, &in[j * 240 + i]);
	    if (*pal > MAXCOLORS) {
		printf("Error: too many colors.\n");
		exit(1);
	    }
	    *pal = updatepalette(palette, *pal, &in[j * 240 + i + 1]);
	    if (*pal > MAXCOLORS) {
		printf("Error: too many colors.\n");
		exit(1);
	    }

	    font[pos].raw[j * 4 + i / 2] =
		frompalette(palette, *pal, &in[j * 240 + i]) |
		(frompalette(palette, *pal, &in[j * 240 + i + 1]) << 4);
	}
    }
    for (temp = 0; temp < pos; temp++) {
	if (memcmp(font[pos].raw, font[temp].raw, sizeof(agb_font)) == 0) {
	    *screen = temp + ADDER;
	    return pos;
	}
	/* Also check mirroring */
	h = horizontal_mirror(&font[pos]);
	if (memcmp(h->raw, font[temp].raw, sizeof(agb_font)) == 0) {
	    free(h);
	    *screen = (temp | (1 << 10)) + ADDER;
	    return pos;
	}
	free(h);
	v = vertical_mirror(&font[pos]);
	if (memcmp(v->raw, font[temp].raw, sizeof(agb_font)) == 0) {
	    free(v);
	    *screen = (temp | (1 << 11)) + ADDER;
	    return pos;
	}
	vh = horizontal_mirror(v);
	free(v);
	if (memcmp(vh->raw, font[temp].raw, sizeof(agb_font)) == 0) {
	    free(vh);
	    *screen = (temp | (1 << 10) | (1 << 11)) + ADDER;
	    return pos;
	}
	free(vh);
    }
    *screen = pos + ADDER;
    return pos + 1;
}

int main(int argc, char **argv)
{
    int width, height;
    u8 pos, pal;
    int x, y, a, out;
    unsigned int cfont_size, cscreen_size, total_size, count;
    Imlib_Image image;
    DATA32 *in;
    char buffer[24];
    u16 screen[32 * 20];
    unsigned char compressed_screen[32 * 20 * 4];
    agb_font font[MAXFONTS + 1];
    unsigned char compressed_font[(MAXFONTS + 1) * sizeof(agb_font) * 2];
    unsigned short palette[MAXCOLORS + 1];

    memset(palette,0,32);
    memset(screen,0,32*20*2);

    if (argc >= 2 && !(argc&1)) {
	out = creat(argv[1], 0600);
	write(out, &borderid, 4);
	count = (argc - 2)>>1;
	write(out, &count, 4);
	for (a = 2; a < argc; a+= 2) {
	    image = imlib_load_image(argv[a]);
	    if (!image) {
		printf("Invalid image: %s. Aborting.\n", argv[a]);
		return 1;
	    }
	    imlib_context_set_image(image);
	    in = imlib_image_get_data_for_reading_only();
	    width = imlib_image_get_width();
	    height = imlib_image_get_height();
	    if (width != 240 || height != 160) {
		printf("Illegal sized image %s (should be 240x160).\n",
		       argv[a]);
		return 1;
	    }

	    pal = 1;
	    pos = 0;

	    // Add non-duplicate fonts.
	    for (y = 0; y < 160; y += 8) {
		for (x = 0; x < 240; x += 8) {
		    // Don't add blank center
		    if (y >= 8 && y < 152 && x >= 40 && x < 200)
			screen[y / 8 * 32 + x / 8] = 0x0120;
		    else {
			pos =
			    updatefont(font, pos, palette, &pal,
				       &screen[y / 8 * 32 + x / 8],
				       &in[y * 240 + x]);
			if (pos >= MAXFONTS) {
			    printf
				("Error: too many tiles needed for %s.\n",
				 argv[a]);
			    return 1;
			}
		    }
		}
	    }
	    imlib_free_image();
	    memset(buffer, 0, 24);
	    strncpy(buffer, argv[a+1], 23);
	    cfont_size = LZ_Compress((u8 *) font, compressed_font, pos * 32) + 4;
	    cscreen_size = LZ_Compress((u8 *) screen, compressed_screen, 20 * 32 * 2);
	    total_size = cfont_size + cscreen_size + 24 + 4 + 32;
	    write(out, &total_size, 4);
	    write(out, &buffer, 24);
	    write(out, &cfont_size, 4);
	    write(out, compressed_font, cfont_size-4);
	    write(out, palette, 32);
	    write(out, compressed_screen, cscreen_size);
	}
	return 0;
    } else {
	printf("Usage: %s <out_bound> <image1> <image1 title> ...\n", argv[0]);
	return 1;
    }
}
